/* fifo.h - C++ pelda tobbszalu FIFO template osztalyra. */

#ifndef FIFO_H
#define FIFO_H

#include <pthread.h>
#include <limits.h>

/* A tarolt elemek maximalis szama */
#define MAX_ELEMENTS INT_MAX

class Lock
{
	pthread_mutex_t* mutex;	
public:
	Lock(pthread_mutex_t* mutex):mutex(mutex){pthread_mutex_lock(mutex);}
	~Lock(){pthread_mutex_unlock(mutex);}
};

/* FIFO template osztaly */
template<class Type> class Fifo
{
protected:
  Type* data;
  int elements;
  pthread_mutex_t critical_section_mutex;

public:
  Fifo();
  Fifo(Fifo&);      /* Masolo konstruktor */
  ~Fifo();
  bool isEmpty();   /* Ha a FIFO ures, akkor igazzal ter vissza */
  int push(Type);   /* Uj elemet helyez el a FIFO-ban */
  int  pop(Type&);  /* Kivesz egy elemet a FIFO-bol */
  void clear();     /* Kitorli a FIFO tartalmat */
 };

/* Alapertelmezett konstruktor */
template<class Type>
Fifo<Type>::Fifo()
{
  /* Inicializaljuk a mutex-et es a tarolo strukturakat. */
  pthread_mutex_init(&critical_section_mutex,NULL);
  data=NULL;
  elements=0;
}

/* Masolo konstruktor */
template<class Type>
Fifo<Type>::Fifo(Fifo<Type>&fifo)
{
  /* Inicializaljuk a mutex-et. */
  pthread_mutex_init(&critical_section_mutex,NULL);

  /* Lefoglaljuk a masolando FIFO belso adatstrukturajahoz a hozzaferest. */
  Lock lock(&fifo.critical_section_mutex);

  /* Kimasoljuk az adatokat. */
  elements=fifo.elements;
  data=new Type[elements];
  for(int i=0;i<elements;i++)
  {
    data[i]=fifo.data[i];
  }

  /* Felszabaditjuk a masolando FIFO mutex-et. */
}

/* Destruktor */
template<class Type>
Fifo<Type>::~Fifo()
{
  	
  /* Felszabaditjuk az adatteruleteket. */
  if(data!=NULL) delete[] data;

  pthread_mutex_destroy(&critical_section_mutex);
}

/* Vizsgalja a FIFO allapotat. Ha a FIFO ures, akkor igaz ertekkel ter
   vissza. */
template<class Type>
bool Fifo<Type>::isEmpty()
{
  bool isempty=false
;

  /* Lefoglaljuk a mutex-et. */
  Lock lock(&critical_section_mutex);

  /* A vedett szakaszban megvizsgaljuk az erteket. */
  isempty=(elements==0);

  /* Visszaterunk a kiolvasott ertekkel. */
  return isempty;
}

/* Uj elemet helyez el a FIFO-ban. Sikeres vegrehajtas eseten a visszateresi
   ertek 0. */
template<class Type>
int Fifo<Type>::push(Type e)
{
  /* Lefoglaljuk a mutex-et. */
  Lock lock(&critical_section_mutex);

  /* Megvizsgaljuk, hogy van-e hely az uj elem szamara. */
  if(elements>=MAX_ELEMENTS )
  {
    /* Hiba eseten felszabaditjuk a mutex-et es visszaterunk. */
    return -1;  
  }
  
  /* Memoria teruletet allokalunk, amelyben a megnovekedett tombot
     elhelyezhetjuk. */
  Type* ptr=new Type[elements+1];
  if(ptr==NULL)
  {
    return -1;
  }

  /* Lemasoljuk a regi adatokat az uj memoria teruletre. */
  for(int i=0;i<elements;i++)
  {
    ptr[i]=data[i];
  }

  /* Az adatterulet vegere elhelyezzuk az uj elemet. */
  ptr[elements]=e;

  /* Felszabaditjuk a regi memoria teruletet. */
  if(data!=NULL) delete[] data;

  /* Az uj memoria teruletet hasznaljuk mostantol. */
  data=ptr;

  /* Megnoveljuk az elemek szamat. */
  elements++;

  return 0;
}

/* Kivesz egy elemet a FIFO-bol. Sikeres vegrehajtas eseten a visszateresi
   ertek 0. */
template <class Type>
int Fifo<Type>::pop(Type& ret)
{
  /* Levedjuk a kritikus muveleteket. */
  Lock lock(&critical_section_mutex);

  /* Megvizsgaljuk van-e varakozo elem a FIFO-ban. */
  if(!elements)
  {
    return -1;
  }

  /* A visszaadando a sor elejen allo elem. */
  ret=data[0];

  /* Amennyiben ez volt az utolso adat, akkor a teljes belso tarolo strukturat
     felszabaditjuk es visszaterunk. */
  if(elements==1)
  {
    delete[] data;
    data=NULL;
    elements--;
    return 0;
  }

  /* Ha vannak tovabbi varakozo adatok, akkor uj teruletet allokalunk, amelyben
     a maradek adatokat helyezzuk el. */
  Type*ptr=new Type[elements-1];
  if(ptr==NULL)
  {
    return -1;
  }

  /* Lemasoljuk az elemeket az elso kivetelevel az uj memoria teruletre. */
  for(int i=1;i<elements;i++)
  {
    ptr[i-1]=data[i];
  }

  /* Felszabaditjuk a regi memoria teruletet. */
  delete[] data;

  /* Az uj memoria teruletet hasznaljuk mostantol. */
  data=ptr;

  /* Csokkentjuk az elemek szamat. */
  elements--;

  return 0;
}

/* Kitorli a FIFO tartalmat. */
template<class Type>
void Fifo<Type>::clear()
{
  /* Lefoglaljuk a mutex-et. */
  Lock lock(&critical_section_mutex);

  /* Toroljuk a tarolt adatokat. */
  if(data!=NULL) delete[] data;
  data=NULL;
  elements=0;
}

#endif      /* FIFO_H */
